(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 4.0,
MathReader 4.0, or any compatible application. The data for the notebook 
starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     29799,        811]*)
(*NotebookOutlinePosition[     30481,        835]*)
(*  CellTagsIndexPosition[     30437,        831]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction to Wavelets for Image Compression ", "Title"],

Cell["\<\
Aleks Jakulin (jakulin@acm.org)
December 2001, June 2002\
\>", "Subtitle"],

Cell["\<\
Please note that this note doesn't cover all the features, rather an \
extremely limited (but important) subset of them, mostly related to the \
basics of wavelet transforms and quantization. I still hope that it will be \
useful and educational as an introduction to applying wavelets to image \
compression and denoising.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Basic 1D wavelets", "Section"],

Cell[BoxData[{
    \(\(<< Graphics`Graphics`;\)\), "\[IndentingNewLine]", 
    \(\(<< LinearAlgebra`MatrixManipulation`;\)\)}], "Input"],

Cell["\<\
This is the Cohen-Daubechies-Feauveau  9/7 wavelet filter, the standard \
transform in the JPEG-2000 standard. We implement it with an efficient \
lifting transformation.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(\[Alpha] = \(-1.586134342\);\)\), "\n", 
    \(\(\[Beta] = \(-0.052980118\);\)\), "\n", 
    \(\(\[Gamma] = 0.882911075;\)\), "\n", 
    \(\(\[Delta] = 0.443506852;\)\), "\n", 
    \(\(K = 1.230174105;\)\), "\n", 
    \(\(DWT\  = 
        Compile[{{X, _Real, 1}}, \ 
          Module[{n, N = Length[X]}, Y = X; 
            For[n = 0, 2  n + 3 <= N, 
              Y[\([2  n + 2]\)] = 
                X[\([2  n + 2]\)] + \[Alpha] \((X[\([2  n + 1]\)] + 
                        X[\([2  n + 
                              3]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 1, 2  n + 2 <= N, 
              Y[\([2  n + 1]\)] = 
                X[\([2  n + 1]\)] + \[Beta] \((Y[\([2  n]\)] + 
                        Y[\([2  n + 
                              2]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 0, 2  n + 3 <= N, 
              Y[\([2  n + 2]\)] = 
                Y[\([2  n + 2]\)] + \[Gamma] \((Y[\([2  n + 1]\)] + 
                        Y[\([2  n + 
                              3]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 1, 2  n + 2 <= N, 
              Y[\([2  n + 1]\)] = 
                Y[\([2  n + 1]\)] + \[Delta] \((Y[\([2  n]\)] + 
                        Y[\([2  n + 
                              2]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 0, 2  n + 2 <= N, 
              Y[\([2  n + 
                      2]\)] = \(-K\)\ \ Y[\([2  n + 
                        2]\)]; \(n++\)]; \[IndentingNewLine]For[n = 0, 
              2  n + 1 <= N, 
              Y[\([2  n + 1]\)] = \((1/K)\) 
                  Y[\([2  n + 1]\)]; \(n++\)]; \[IndentingNewLine]Return[
              Y]]];\)\), "\n", 
    \(\(IDWT\  = 
        Compile[{{Y, _Real, 1}}, \ 
          Module[{n, N = Length[Y]}, X = Y; 
            For[n = 0, 2  n + 1 <= N, 
              X[\([2  n + 1]\)] = 
                K\ Y[\([2  n + 1]\)]; \(n++\)]; \[IndentingNewLine]For[n = 0, 
              2  n + 2 <= N, 
              X[\([2  n + 
                      2]\)] = \(-\((1/K)\)\)\ Y[\([2  n + 
                        2]\)]; \(n++\)]; \[IndentingNewLine]For[n = 1, 
              2  n + 2 <= N, 
              X[\([2  n + 1]\)] = 
                X[\([2  n + 1]\)] - \[Delta] \((X[\([2  n]\)] + 
                        X[\([2  n + 
                              2]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 0, 2  n + 3 <= N, 
              X[\([2  n + 2]\)] = 
                X[\([2  n + 2]\)] - \[Gamma] \((X[\([2  n + 1]\)] + 
                        X[\([2  n + 
                              3]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 1, 2  n + 2 <= N, 
              X[\([2  n + 1]\)] = 
                X[\([2  n + 1]\)] - \[Beta] \((X[\([2  n]\)] + 
                        X[\([2  n + 
                              2]\)])\); \(n++\)]; \[IndentingNewLine]For[
              n = 0, 2  n + 3 <= N, 
              X[\([2  n + 2]\)] = 
                X[\([2  n + 2]\)] - \[Alpha] \((X[\([2  n + 1]\)] + 
                        X[\([2  n + 3]\)])\); \(n++\)]; 
            Return[X]]];\)\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(IDWT\)\" is similar to \
existing symbol \"\!\(DWT\)\"."\)], "Message"]
}, Open  ]],

Cell["\<\
Vectors are extended to allow proper transformation with filters that \
reference the surrounding area of every sample. The extension is done with 4 \
elements, as required by the above 9/7 filter.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(VectorExtend\  = 
        Compile[{{X, _Real, 1}}, \ 
          Join[Join[Reverse[Drop[Take[X, 5], 1]], X], 
            Reverse[Drop[Take[X, \(-5\)], \(-1\)]]]];\)\), "\n", 
    \(\(VectorShrink\  = 
        Compile[{{X, _Real, 1}}, \ 
          Drop[Drop[X, 4], \(-4\)]];\)\), "\[IndentingNewLine]", 
    \(\(ExtDWT[X_] := 
        VectorShrink[DWT[VectorExtend[X]]];\)\), "\[IndentingNewLine]", 
    \(\(ExtIDWT[X_] := VectorShrink[IDWT[VectorExtend[X]]];\)\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(ExtIDWT\)\" is similar \
to existing symbol \"\!\(ExtDWT\)\"."\)], "Message"]
}, Open  ]],

Cell["\<\
Note that the transformation is reversible regardless of the vector extension\
\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ExtIDWT[ExtDWT[{1, 2, 3, 4, 5, 6, 7, 8}]]\)], "Input"],

Cell[BoxData[
    \({0.9999999999999999`, 2.`, 3.`, 4.`, 5.`, 6.`, 7.`, 8.`}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(IDWT[DWT[{1, 2, 3, 4, 5, 6, 7, 8}]]\)], "Input"],

Cell[BoxData[
    \({1.`, 2.`, 3.`, 4.`, 5.`, 6.000000000000002`, 7.000000000000002`, 
      8.`}\)], "Output"]
}, Open  ]],

Cell["\<\
But note that the coefficient magnitudes are lower with vector extension\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ExtDWT[{1, 2, 3, 4, 5, 6, 7, 8}]\)], "Input"],

Cell[BoxData[
    \({1.3336405054605858`, \(-0.24999998336082443`\), 3.0732667734702805`, 
      2.7577721735413305`*^-8, 4.946502455076224`, 0.18254356551703407`, 
      7.06341042466827`, \(-0.8650870084380856`\)}\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(DWT[{1, 2, 3, 4, 5, 6, 7, 8}]\)], "Input"],

Cell[BoxData[
    \({0.8128930660591333`, 0.2500000109385473`, 2.9267331896609825`, 
      2.7577721735413305`*^-8, 4.572019855677429`, 1.460348234570178`, 
      8.363243925638098`, \(-9.84139284`\)}\)], "Output"]
}, Open  ]],

Cell["\<\
The high-pass and low-pass subbands are interleaved after the transform, but \
it is often desirable to separate and deinterleave them for further \
processing.\
\>", "Text"],

Cell[BoxData[{
    \(\(DeInterleave = 
        Compile[{{X, _Real, 1}}, 
          Join[First /@ Partition[X, 2], 
            Last\  /@ \ Partition[X, 2]]];\)\), "\n", 
    \(\(Interleave = 
        Compile[{{X, _Real, 1}}, 
          Flatten[Transpose[
              Partition[X, Floor[0.5*Length[X]]]]]];\)\)}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Wavelets on images", "Section"],

Cell["\<\
Wavelet transform is especially useful for transforming images. For this, we \
apply it twice according to the JPEG-2000 standard: first on columns, second \
on rows. Upon this, we deinterleave the image matrix, and possibly \
recursively transform each subband individually further.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ImgDWT = \ 
        Compile[{{X, _Real, 2}}, 
          DeInterleave[
            ExtDWT\  /@ \ 
              Transpose[
                DeInterleave[
                  ExtDWT\  /@ \ 
                    Transpose[\((X - 
                          128)\)]]]]];\)\), "\[IndentingNewLine]", 
    \(\(ImgIDWT = 
        Compile[{{Y, _Real, 2}}, 
          Transpose[
              ExtIDWT\  /@ \ 
                Interleave[Transpose[ExtIDWT\  /@ \ Interleave[Y]]]] + 
            128];\)\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(ImgIDWT\)\" is similar \
to existing symbol \"\!\(ImgDWT\)\"."\)], "Message"]
}, Open  ]],

Cell[BoxData[{
    \(\(Rescale[
          X_] := \((X - Min[X])\)/\((Max[X] - 
              Min[X])\);\)\), "\[IndentingNewLine]", 
    \(\(RescalePlain[X_] := \((X + 128)\)/255;\)\)}], "Input"],

Cell["\<\
We can now import the standard benchmark picture, sized 512x512:\
\>", "Text"],

Cell[BoxData[{
    \(\(m\  = \ 
        Import["\<C:\PROJ\WAVELETS\lenna.pgm\>"];\)\), "\[IndentingNewLine]", \

    \(\(Show[m, ImageSize \[Rule] \ {512 + 32, 512 + 32}];\)\)}], "Input"],

Cell["\<\
We will only work with a small but diverse part of the image, combining sharp \
features, smooth gradients, and complex textures. Note that we conver the \
unsigned byte values [0,256) to signed on the interval [-128,128). This way \
we can exploit gradient symmetry in quantization\
\>", "Text"],

Cell[BoxData[
    \(\(sub\  = \ 
        SubMatrix[m[\([1, 1]\)], {170, 178}, {128, 128}] - 128;\)\)], "Input"],

Cell[BoxData[
    \(\(Show[Graphics[Raster[\((sub + 128)\)/255]], 
        AspectRatio \[Rule] 1];\)\)], "Input"],

Cell["\<\
By transposing the image, we can easily implement the 2D DWT:\
\>", "Text"],

Cell[BoxData[{
    \(\(ImgDWT = 
        Compile[{{X, _Real, 2}}, 
          ExtDWT\  /@ \ 
            Transpose[
              ExtDWT\  /@ \ Transpose[X]]];\)\), "\[IndentingNewLine]", 
    \(\(ImgIDWT = 
        Compile[{{X, _Real, 2}}, 
          ExtIDWT /@ Transpose[ExtIDWT /@ X]];\)\)}], "Input"],

Cell["\<\
It is very useful to interleave and deinterlave images too. We do this by \
creating four subimages, for the HH, LH, HL, and LL subbands.\
\>", "Text"],

Cell[BoxData[{
    \(\(DeInterleave2[X_]\  := \ 
        Table[Table[
            Table[X[\([2  a - i, 2  b - j]\)], {a, 
                1, \(Dimensions[X]\)[\([1]\)]/2}, {b, 
                1, \(Dimensions[X]\)[\([2]\)]/2}], {i, 1, 0, \(-1\)}], {j, 1, 
            0, \(-1\)}];\)\), "\[IndentingNewLine]", 
    \(Interleave2[X_] := 
      Flatten[Table[{Flatten[
              Table[{X[\([1, 1, a, b]\)], X[\([1, 2, a, b]\)]}, {a, 
                  1, \(Dimensions[X]\)[\([3]\)]}]], 
            Flatten[Table[{X[\([2, 1, a, b]\)], X[\([2, 2, a, b]\)]}, {a, 
                  1, \(Dimensions[X]\)[\([3]\)]}]]}, {b, 
            1, \(Dimensions[X]\)[\([4]\)]}], 1]\)}], "Input"],

Cell["\<\
Now that we have everything ready, we can embark on processing the chosen \
section of the image.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(isub\  = \ ImgDWT[sub];\)\), "\[IndentingNewLine]", 
    \(\(decomp\  = \ DeInterleave2[isub];\)\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(isub\)\" is similar to \
existing symbol \"\!\(sub\)\"."\)], "Message"]
}, Open  ]],

Cell["Let's show all four subbands:", "Text"],

Cell[BoxData[
    \(\(Show[
        GraphicsArray[
          Map[Graphics[Raster[\((# + 128)\)/256], AspectRatio \[Rule] 1] &, 
            decomp, {2}], GraphicsSpacing \[Rule] 0]];\)\)], "Input"],

Cell["The transformation is reversible, as we now show.", "Text"],

Cell[BoxData[
    \(\(Show[Graphics[Raster[\((ImgIDWT[Interleave2[decomp]] + 128)\)/255]], 
        AspectRatio \[Rule] 1];\)\)], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Denoising", "Section"],

Cell["\<\
Now, let's try to denoise the image, by chopping near-zero values in the LH, \
HL and HH components (or low-power white noise), and subsequently \
reconstructing the image. Note that we chop away more in the HH subband than \
in the LH and HL - but we didn't have to. This technique is called hard  \
thresholding.\
\>", "Text"],

Cell[BoxData[{
    \(\(thdecomp\  = \ decomp;\)\), "\[IndentingNewLine]", 
    \(\(thdecomp[\([2, 2]\)]\  = \ 
        Chop[decomp[\([2, 2]\)], 32];\)\), "\[IndentingNewLine]", 
    \(\(thdecomp[\([1, 2]\)]\  = \ 
        Chop[decomp[\([1, 2]\)], 8];\)\), "\[IndentingNewLine]", 
    \(\(thdecomp[\([2, 1]\)]\  = \ Chop[decomp[\([2, 1]\)], 8];\)\)}], "Input"],

Cell[BoxData[
    \(\(Show[
        Graphics[Raster[\((ImgIDWT[Interleave2[thdecomp]] + 128)\)/255]], 
        AspectRatio \[Rule] 1];\)\)], "Input"],

Cell["\<\
Soft thresholding shrinks the magnitude of the coefficient values outside the \
threshold, but it's worse in our example.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(SoftThresh[
          N_] := \ {\((X_?NumberQ \[RuleDelayed] 
              0\  /; \ Abs[X] \[LessEqual] \ N)\), \((X_?
                NumberQ \[RuleDelayed] \ 
              X - N\  /; \ Abs[X] > N)\)};\)\), "\[IndentingNewLine]", 
    \(\(shdecomp\  = \ decomp;\)\), "\[IndentingNewLine]", 
    \(\(shdecomp[\([2, 2]\)]\  = \ 
        decomp[\([2, 2]\)] /. SoftThresh[32];\)\), "\[IndentingNewLine]", 
    \(\(shdecomp[\([1, 2]\)]\  = \ 
        decomp[\([1, 2]\)] /. SoftThresh[8];\)\), "\[IndentingNewLine]", 
    \(\(shdecomp[\([2, 1]\)]\  = \ 
        decomp[\([2, 1]\)] /. SoftThresh[8];\)\), "\[IndentingNewLine]", 
    \(\(Show[
        Graphics[Raster[\((ImgIDWT[Interleave2[shdecomp]] + 128)\)/255]], 
        AspectRatio \[Rule] 1];\)\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(shdecomp\)\" is similar \
to existing symbol \"\!\(thdecomp\)\"."\)], "Message"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Quantization", "Section"],

Cell["\<\
Image compression is based on a slightly different concept - quantization. \
Instead of having 256 levels of grey, we might have only 16. Such an image \
would look bad. But we can do just this on the LH, HL and HH subbands, \
without losing much. All we're going to do is rounding - with different \
amounts depending on the subband. This is called deadzone (rounding down) \
uniform (division by a number) scalar (it operates on individual color \
components of individual pixels which are scalars) quantization. Let's see:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Quantize[N_] := \ 
        X_?NumberQ \[RuleDelayed] \((Round[X/N])\)*
            N;\)\), "\[IndentingNewLine]", 
    \(\(qdecomp\  = \ decomp;\)\), "\[IndentingNewLine]", 
    \(\(qdecomp[\([2, 2]\)]\  = \ 
        decomp[\([2, 2]\)] /. Quantize[64];\)\), "\[IndentingNewLine]", 
    \(\(qdecomp[\([1, 2]\)]\  = \ 
        decomp[\([1, 2]\)] /. Quantize[32];\)\), "\[IndentingNewLine]", 
    \(\(qdecomp[\([2, 1]\)]\  = \ 
        decomp[\([2, 1]\)] /. Quantize[32];\)\), "\[IndentingNewLine]", 
    \(\(Show[
        GraphicsArray[{Graphics[
              Raster[\((ImgIDWT[Interleave2[qdecomp]] + 128)\)/255], 
              AspectRatio \[Rule] 1], 
            Graphics[Raster[\((sub + 128)\)/255], 
              AspectRatio \[Rule] 1]}]];\)\[IndentingNewLine]\)}], "Input"],

Cell[BoxData[
    \(General::"spell1" \(\(:\)\(\ \)\) 
      "Possible spelling error: new symbol name \"\!\(qdecomp\)\" is similar \
to existing symbol \"\!\(decomp\)\"."\)], "Message"]
}, Open  ]],

Cell["\<\
As you can see,  not that much quality was lost - especially considering that \
the above picture is enlarged on the screen! But the final filesize due to \
skipping  bits is (compare with 100%):\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(N[25 + \((50*4/8)\) + \((25*3/8)\)]\)], "Input"],

Cell[BoxData[
    \(59.375`\)], "Output"]
}, Open  ]],

Cell["\<\
This might not seem much, but note that the remaining information is much \
more (i.e. orders of magnitude more) compressible with ordinary entropy \
coding techniques (usually preceded by run-length coding), as we also \
denoised the image! Furthermore, we can recursively filter the LL subband \
again, and repeat the procedure on LLLH, LLHL, and LLHH.

Examine the subbands after quantization. Note that the LH, HL and HH subbands \
have become extremely simple and very well compressible.\
\>", "Text"],

Cell[BoxData[
    \(\(Show[
        GraphicsArray[
          Map[Graphics[Raster[\((# + 128)\)/256], 
                AspectRatio \[Rule] 1] &, \((qdecomp)\), {2}], 
          GraphicsSpacing \[Rule] 0]];\)\)], "Input"],

Cell["\<\
We notice that the sharp edge information remains in the LH, HL and HH \
subbands, but most of the lower amplitude noise, as well as the exact \
magnitudes of the edges have been compromised.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Other decompositions", "Section"],

Cell["\<\
Let us automatize some earlier operations. Specifically,we will perform DWT \
recursively in each subband. This is sometimes called a packet or SPACL \
decomposition (if we recursively subdivided the LLLL once or twice more), \
unlike the above 1-level Mallat decompositions where the high-pass subbands \
remain undecomposed. The decompositions can be performed recursively - if we \
performed a Mallat decomposition again on the LL, we would obtain a 2-level \
Mallat decomposition. The decomposition below is actually a 2-level packet \
decomposition, as we performed it twice recursively, and it's packet \
decomposition because it was performed on all subbands.\
\>", "Text"],

Cell[BoxData[
    \(\(trecomp = Map[DeInterleave2[ImgDWT[#]] &, decomp, {2}];\)\)], "Input"],

Cell[BoxData[{
    \(\(GfxMat[X_, s_] := 
        GraphicsArray[
          Map[GraphicsArray[
                Map[Graphics[Raster[\((# + \((s/2)\))\)/s], 
                      AspectRatio \[Rule] 1] &, #, {2}], 
                GraphicsSpacing \[Rule] 0] &, X, {2}], 
          GraphicsSpacing \[Rule] 0];\)\), "\[IndentingNewLine]", 
    \(\(Show[GfxMat[trecomp, 256]];\)\)}], "Input"],

Cell["\<\
We can see that the LH*, HL* and HH* subbands appear to be quite \
uninformative. It's time to try to understand what wavelet transform is \
really about. The low-pass filter is a kind of a predictor: what can we say \
about the average value of a group of pixels. The high-pass filters, on the \
other hand, store the errors we make with the low-pass predictors. LH and HL \
store the errors we make when we try to get to average values of \
column-errors across rows, or vice versa, whereas HH stores LH's and HL's \
errors. 

We see that the LL pass over LH,HL,HH subbands is tending towards the average \
of 0. This is a good indicator of noise as those irregularities that appear \
merely in the highest resolution, but cannot be generalized to a wider area. \
However, we must note that people do recognize the absence of noise. The main \
difference between the original and the compressed image is that the \
compressed one is too smooth.

In the following experiments, we will try to capture the regularities of \
noise. To do this, we will need a different kind of a transform.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Describing quantization noise", "Section"],

Cell["\<\
Let's examine the error introduced by quantization. We notice that the error \
is pretty much random, with the exception of a few features in the HH band. \
The idea we will now pursue is to try to describe this error and re-introduce \
it in the decompressed image.\
\>", "Text"],

Cell[BoxData[
    \(mcomp = decomp - qdecomp; 
    Show[GraphicsArray[
        Map[Graphics[Raster[\((# + 128)\)/256], AspectRatio \[Rule] 1] &, 
          mcomp, {2}], GraphicsSpacing \[Rule] 0]];\)], "Input"],

Cell["Let us now examine the frequency distributions of error:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(DisplayTogetherArray[{{Histogram[Flatten[\(mcomp[\([1]\)]\)[\([1]\)]]], 
          Histogram[Flatten[\(mcomp[\([1]\)]\)[\([2]\)]]]}, {Histogram[
            Flatten[\(mcomp[\([2]\)]\)[\([1]\)]]], 
          Histogram[Flatten[\(mcomp[\([2]\)]\)[\([2]\)]]]}}]\)], "Input"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  GraphicsArray  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["\<\
The LH and HL band distributions are narrower because of the lower level of \
quantization, and are not very Gaussian: this is because noise got through \
quantization in those two bands.  We could have been more radical when \
quantizing - disposing with more than 4 bits without losing much information. \
But not to dawdle, let's now capture the noise (assuming it is Gaussian) and \
reintroduce it in the image. 

 There is an important consideration to be made: a part of the quantization \
noise is indeed noise, but a part is lost information. The lost information \
shouldn't be considered a part of noise - as we cannot reintroduce it. Thus \
not to overestimate the noise , we divide the estimated std. deviation by \
2.\
\>", "Text"],

Cell[BoxData[{
    \(<< Statistics`NormalDistribution`\), "\[IndentingNewLine]", 
    \(\(NoiseProp[array_] := {Mean[Flatten[array]], 
          StandardDeviation[Flatten[array]]/2};\)\), "\[IndentingNewLine]", 
    \(\(Noise[{mean_, stddev_}, dim_] := 
        Partition[RandomArray[NormalDistribution[mean, stddev], dim*dim], 
          dim];\)\)}], "Input"],

Cell["\<\
The noise description is very compact, and adds very little to the compressed \
file size. Especially considering the fact that the mean is always \
approximately 0, and does not need to be stored.\
\>", "Text"],

Cell[BoxData[
    \(\(noisedesc = Map[NoiseProp, decomp - qdecomp, {2}];\)\)], "Input"],

Cell["Here we add the noise to the compressed image:", "Text"],

Cell[BoxData[
    \(\(enhancedqdecomp\  = 
        Table[qdecomp[\([i, j]\)] + 
            Noise[noisedesc[\([i, j]\)], \(Dimensions[
                  qdecomp[\([i, j]\)]]\)[\([1]\)]], {i, 
            1, \(Dimensions[qdecomp]\)[\([1]\)]}, {j, 
            1, \(Dimensions[qdecomp]\)[\([2]\)]}];\)\)], "Input"],

Cell["\<\
And now we compare the \"en-noised\" compressed image to the original image \
and the normal \"de-noised\" compressed image:\
\>", "Text"],

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{Graphics[
              Raster[\((ImgIDWT[Interleave2[qdecomp]] + 128)\)/255], 
              AspectRatio \[Rule] 1], 
            Graphics[Raster[\((sub + 128)\)/255], AspectRatio \[Rule] 1], 
            Graphics[
              Raster[\((ImgIDWT[Interleave2[enhancedqdecomp]] + 128)\)/255], 
              AspectRatio \[Rule] 1]}]];\)\)], "Input"],

Cell["\<\
Although the en-noised image does look better than the de-noised one, we note \
that there is more noise than  there was in the beginning. This is partly \
because we equated some information lost in quantization with noise, and \
partly because there are noiser and less noisy areas on the image. For \
example, the skin as a less noisy part of the image has received more noise \
than there was in the beginning, whereas the edges and sharp features \
received too little. We corrected this by introducing the correction factor \
of 2 (we divided the standard deviation by 2) which makes the skin \
appropriately smooth, but the hat feathers excessively smooth.\
\>", "Text"],

Cell["\<\
We would have achieved better results by decomposing in the spatial domain. \
For example, the texture of the hat is noisier than the texture of the skin, \
so the noise distribution should be computed differently there. There is \
another important but yet unsolved problem: a part of the quantization noise \
is indeed noise, but a part is lost information. The lost information \
shouldn't be considered a part of noise - as we cannot reintroduce it.\
\>", "Text"],

Cell[BoxData[
    \(\(diff\  = \ ImgIDWT[Interleave2[qdecomp]] - sub;\)\)], "Input"],

Cell[TextData[StyleBox["TODOTODO (Miha): kaj ce bi sum dodali le na vrhu \
navadne slike?",
  FontWeight->"Bold"]], "Text"],

Cell[TextData[StyleBox["TODOTODO(Aleks): locevanje suma od kvantizacijskih \
artifaktov: sum je gaussov, kvantizacijski artifakti pa niso (uniformna \
porazdelitev?)",
  FontWeight->"Bold"]], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Variance map", "Section"],

Cell["\<\
Now, we compute the standard deviations of quantization residuals at \
different resolutions. We will call these \"variance maps\".\
\>", "Text"],

Cell["\<\
We will need a routine which will bunch together neighboring pixels\
\>", "Text"],

Cell[BoxData[{
    \(\(Whack[M_, res_] := 
        Map[Map[Flatten, #] &, 
          Transpose[
            Map[Partition[#, res] &, Transpose[Map[Partition[#, res] &, M]], 
              2]]];\)\), "\[IndentingNewLine]", 
    \(FiltO[x_] := 0 /; x != 0; \ FiltO[x_] := 1 /; x == 0; 
    Filter[C_, res_] := 
      Map[FiltO, 
        Map[Map[Flatten, #] &, 
          Transpose[
            Map[Partition[#, res] &, Transpose[Map[Partition[#, res] &, C]], 
              2]]], {3}];\), "\[IndentingNewLine]", 
    \(\(VarianceMap[{M_, C_}, res_] := 
        Map[StandardDeviation, 
          Whack[M, res]*Filter[C, res], {2}];\)\)}], "Input"],

Cell[BoxData[
    \(\(Show[
        GraphicsArray[
          Map[Block[{res = #}, 
                GraphicsArray[
                  Map[Graphics[Raster[\((VarianceMap[#, res] + 16)\)/32], 
                        AspectRatio \[Rule] 1] &, 
                    Table[{mcomp[\([i, j]\)], qdecomp[\([i, j]\)]}, {i, 1, 
                        2}, {j, 1, 2}], {2}], 
                  GraphicsSpacing \[Rule] 0]] &, {2, 4, 8, 16, 32, 64}], 
          GraphicsSpacing \[Rule] 0]];\)\)], "Input"],

Cell["\<\
It seems that the variance map at 16 pixel resolution works best: not too \
much detail, yet not too little of it. Let's now apply variance map at \
resolution 16 for noise reconstruction. We divide the standard deviations by \
2, or else the amount of noise is overestimated. It is important to realize \
that only noise must be reintroduced, but not just noise is lost in \
quantization.\
\>", "Text"],

Cell[BoxData[{
    \(\(res\  = \ 16;\)\), "\[IndentingNewLine]", 
    \(\(vmap\  = \ 
        Map[VarianceMap[#, res] &, 
          Table[{mcomp[\([i, j]\)], qdecomp[\([i, j]\)]}, {i, 1, 2}, {j, 1, 
              2}], {2}];\)\), "\[IndentingNewLine]", 
    \(\(NoiseRectangle[stddev_, dim_] := 
        Partition[RandomArray[NormalDistribution[0.0, stddev/1.5], dim*dim], 
          dim];\)\)}], "Input"],

Cell[BoxData[
    \(\(noiserects\  = \ 
        Map[NoiseRectangle[#, res] &, vmap, {4}];\)\)], "Input"],

Cell[BoxData[{
    \(\(<< LinearAlgebra`MatrixManipulation`;\)\), "\[IndentingNewLine]", 
    \(MyFold[fun_, array_] := 
      Fold[fun, First[array], Rest[array]]\), "\[IndentingNewLine]", 
    \(Merging[array_] := 
      Flatten[Map[MyFold[AppendRows, #] &, array], 1]\)}], "Input"],

Cell[BoxData[
    \(\(noisemap = Map[Merging, noiserects, {2}];\)\)], "Input"],

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{Graphics[
              Raster[\((ImgIDWT[Interleave2[qdecomp]] + 128)\)/255], 
              AspectRatio \[Rule] 1], 
            Graphics[Raster[\((sub + 128)\)/255], AspectRatio \[Rule] 1], 
            Graphics[
              Raster[\((ImgIDWT[Interleave2[noisemap + qdecomp]] + 128)\)/
                  255], AspectRatio \[Rule] 1]}]];\)\)], "Input"],

Cell["\<\
Although the noise levels on the skin and in the feathers are now more \
correct, there are still visible transitions between different noise levels. \
More troubling, there is too much noise in the areas around sharp edges. \
Further work should address proper computation of noise levels around sharp \
edges.\
\>", "Text"],

Cell[CellGroupData[{

Cell["SPACL decomposition of  residuals:", "Subsection"],

Cell["\<\
We notice that we do not need to compute the standard deviations in such an \
expensive way: it is merely an issue of averaging absolute residuals. We \
might wonder if we could describe the noise better with Spacl decomposition -\
\
\>", "Text"],

Cell[BoxData[{
    \(\(abstrecomp = 
        Map[DeInterleave2[ImgDWT[#]] &, 
          Abs[mcomp], {2}];\)\), "\[IndentingNewLine]", 
    \(\(Show[GfxMat[abstrecomp, 64]];\)\)}], "Input"],

Cell["\<\
One interesting approach would be discarding the noisy high-pass subband of \
variance and quantizing the low-pass subbands of variance.\
\>", "Text"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"4.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 730}},
WindowSize->{1016, 706},
WindowMargins->{{-7, Automatic}, {Automatic, 0}},
CellLabelAutoDelete->True
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1739, 51, 64, 0, 105, "Title"],
Cell[1806, 53, 84, 3, 95, "Subtitle"],
Cell[1893, 58, 347, 6, 52, "Text"],

Cell[CellGroupData[{
Cell[2265, 68, 36, 0, 53, "Section"],
Cell[2304, 70, 136, 2, 50, "Input"],
Cell[2443, 74, 194, 4, 33, "Text"],

Cell[CellGroupData[{
Cell[2662, 82, 3124, 66, 410, "Input"],
Cell[5789, 150, 180, 3, 24, "Message"]
}, Open  ]],
Cell[5984, 156, 221, 4, 33, "Text"],

Cell[CellGroupData[{
Cell[6230, 164, 497, 10, 90, "Input"],
Cell[6730, 176, 186, 3, 24, "Message"]
}, Open  ]],
Cell[6931, 182, 103, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[7059, 189, 74, 1, 30, "Input"],
Cell[7136, 192, 90, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7263, 198, 68, 1, 30, "Input"],
Cell[7334, 201, 111, 2, 29, "Output"]
}, Open  ]],
Cell[7460, 206, 96, 2, 33, "Text"],

Cell[CellGroupData[{
Cell[7581, 212, 65, 1, 30, "Input"],
Cell[7649, 215, 229, 3, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7915, 223, 62, 1, 30, "Input"],
Cell[7980, 226, 214, 3, 29, "Output"]
}, Open  ]],
Cell[8209, 232, 184, 4, 33, "Text"],
Cell[8396, 238, 323, 8, 50, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[8756, 251, 37, 0, 53, "Section"],
Cell[8796, 253, 307, 5, 52, "Text"],

Cell[CellGroupData[{
Cell[9128, 262, 523, 15, 50, "Input"],
Cell[9654, 279, 186, 3, 24, "Message"]
}, Open  ]],
Cell[9855, 285, 195, 4, 50, "Input"],
Cell[10053, 291, 88, 2, 33, "Text"],
Cell[10144, 295, 187, 4, 50, "Input"],
Cell[10334, 301, 306, 5, 52, "Text"],
Cell[10643, 308, 111, 2, 30, "Input"],
Cell[10757, 312, 113, 2, 30, "Input"],
Cell[10873, 316, 85, 2, 33, "Text"],
Cell[10961, 320, 303, 8, 50, "Input"],
Cell[11267, 330, 161, 3, 33, "Text"],
Cell[11431, 335, 682, 13, 110, "Input"],
Cell[12116, 350, 121, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[12262, 357, 133, 2, 50, "Input"],
Cell[12398, 361, 180, 3, 24, "Message"]
}, Open  ]],
Cell[12593, 367, 45, 0, 33, "Text"],
Cell[12641, 369, 197, 4, 30, "Input"],
Cell[12841, 375, 65, 0, 33, "Text"],
Cell[12909, 377, 138, 2, 30, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[13084, 384, 28, 0, 53, "Section"],
Cell[13115, 386, 338, 6, 52, "Text"],
Cell[13456, 394, 359, 6, 90, "Input"],
Cell[13818, 402, 149, 3, 30, "Input"],
Cell[13970, 407, 145, 3, 33, "Text"],

Cell[CellGroupData[{
Cell[14140, 414, 784, 15, 130, "Input"],
Cell[14927, 431, 189, 3, 24, "Message"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[15165, 440, 31, 0, 53, "Section"],
Cell[15199, 442, 549, 8, 71, "Text"],

Cell[CellGroupData[{
Cell[15773, 454, 803, 16, 170, "Input"],
Cell[16579, 472, 186, 3, 24, "Message"]
}, Open  ]],
Cell[16780, 478, 219, 4, 33, "Text"],

Cell[CellGroupData[{
Cell[17024, 486, 68, 1, 30, "Input"],
Cell[17095, 489, 41, 1, 29, "Output"]
}, Open  ]],
Cell[17151, 493, 516, 9, 90, "Text"],
Cell[17670, 504, 219, 5, 30, "Input"],
Cell[17892, 511, 215, 4, 33, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[18144, 520, 39, 0, 53, "Section"],
Cell[18186, 522, 690, 10, 90, "Text"],
Cell[18879, 534, 92, 1, 30, "Input"],
Cell[18974, 537, 387, 8, 90, "Input"],
Cell[19364, 547, 1111, 19, 185, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[20512, 571, 48, 0, 53, "Section"],
Cell[20563, 573, 290, 5, 52, "Text"],
Cell[20856, 580, 210, 4, 30, "Input"],
Cell[21069, 586, 72, 0, 33, "Text"],

Cell[CellGroupData[{
Cell[21166, 590, 290, 4, 50, "Input"],
Cell[21459, 596, 135, 3, 29, "Output"]
}, Open  ]],
Cell[21609, 602, 754, 13, 128, "Text"],
Cell[22366, 617, 360, 6, 70, "Input"],
Cell[22729, 625, 221, 4, 33, "Text"],
Cell[22953, 631, 87, 1, 30, "Input"],
Cell[23043, 634, 62, 0, 33, "Text"],
Cell[23108, 636, 312, 6, 50, "Input"],
Cell[23423, 644, 148, 3, 33, "Text"],
Cell[23574, 649, 400, 8, 70, "Input"],
Cell[23977, 659, 687, 10, 90, "Text"],
Cell[24667, 671, 477, 7, 71, "Text"],
Cell[25147, 680, 84, 1, 30, "Input"],
Cell[25234, 683, 123, 2, 33, "Text"],
Cell[25360, 687, 199, 3, 33, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[25596, 695, 31, 0, 53, "Section"],
Cell[25630, 697, 155, 3, 33, "Text"],
Cell[25788, 702, 91, 2, 33, "Text"],
Cell[25882, 706, 644, 15, 110, "Input"],
Cell[26529, 723, 490, 10, 130, "Input"],
Cell[27022, 735, 413, 7, 52, "Text"],
Cell[27438, 744, 404, 8, 70, "Input"],
Cell[27845, 754, 104, 2, 30, "Input"],
Cell[27952, 758, 284, 5, 70, "Input"],
Cell[28239, 765, 78, 1, 30, "Input"],
Cell[28320, 768, 407, 8, 70, "Input"],
Cell[28730, 778, 335, 6, 52, "Text"],

Cell[CellGroupData[{
Cell[29090, 788, 56, 0, 47, "Subsection"],
Cell[29149, 790, 256, 5, 52, "Text"],
Cell[29408, 797, 188, 4, 50, "Input"],
Cell[29599, 803, 160, 3, 33, "Text"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

